import tensorflow as tf


class Model:

    def __init__(self, data, config):
        self.data = data
        self.config = config

    def create(self):
        return _create(self.data, self.config)


def _create(data, config):
    contextual_inputs, contextual_outputs = \
        _create_contextual_inputs(data, config)
    sequential_inputs, sequential_outputs, _ = \
        _create_sequential_inputs(data, config)
    # Construct the model
    # ...
    # model = tf.keras.Model(
    #     inputs={**contextual_inputs, **sequential_inputs},
    #     outputs=outputs,
    # )
    # model.compile(...)
    # return model


def _create_contextual_inputs(data, config):
    inputs = {
        name: tf.keras.Input(name=name, shape=(), dtype=data.schema[name].kind)
        for name in data.contextual_feature_names
    }
    layer = tf.keras.layers.DenseFeatures(
        feature_columns=data.create_feature_columns('contextual'))
    return inputs, layer(inputs)


def _create_sequential_inputs(data, config):
    inputs = {
        name: tf.keras.Input(name=name,
                             shape=(None,),
                             dtype=data.schema[name].kind,
                             sparse=True)
        for name in data.sequential_feature_names
    }
    layer = tf.keras.experimental.SequenceFeatures(
        feature_columns=data.create_feature_columns('sequential'))
    return (inputs, *layer(inputs))



class Task:

    def run(self, action, config):
        getattr(self, '_run_' + action)(config)

    def _run_preprocessing(self, config):
        from forecast.pipeline import Pipeline
        Pipeline(config).run().wait_until_finish()



import os
import tensorflow as tf
import tensorflow_transform as tft

from forecast.schema import Schema
from forecast.support import list_files


class Data:

    def __init__(self, config):
        # Find the path with the latest preprocessed data by listing all entries
        # matching a given pattern and take the last one, assuming that the
        # alphabetical order is also chronological
        self.path = list_files(config['path'])[-1]
        # Find the path of the latest transforms using the same strategy
        transform_path = list_files(config['transform_path'])[-1]
        # Create a schema according to the configuration file
        self.schema = Schema(config['schema'])
        # Find the names of contextual columns
        self.contextual_names = self.schema.select('contextual')
        # Find the names of sequential columns
        self.sequential_names = self.schema.select('sequential')
        self.modes = config['modes']
        # Load all transforms
        transforms = [mode['transform'] for mode in self.modes.values()]
        self.transforms = {
            name: tft.TFTransformOutput( \
                os.path.join(transform_path, name, 'transform'))
            for name in set(transforms)
        }

    def create(self, name):
        config = self.modes[name]

        def _preprocess_transformed(proto):
            spec = self.transforms[config['transform']] \
                .transformed_feature_spec()
            example = tf.io.parse_single_example(proto, spec)
            return (
                {name: example[name] for name in self.contextual_names},
                {
                    # Convert the sequential columns from sparse to dense
                    name: self.schema[name].to_dense(example[name])
                    for name in self.sequential_names
                },
            )

        def _preprocess_untransformed(proto):
            spec = self.schema.to_feature_spec()
            example = tf.parse_single_example(proto, spec)
            for name in self.contextual_names:
                example[name] = tf.expand_dims(example[name], -1)
            for name in self.sequential_names:
                example[name] = tf.sparse.expand_dims(example[name], -1)
            example = self.transforms[config['transform']] \
                .transform_raw_features(example)
            return (
                {
                    name: tf.reshape(example[name], [-1])
                    for name in self.contextual_names
                },
                {
                    # Convert the sequential columns from sparse to dense
                    name: tf.reshape( \
                        self.schema[name].to_dense(example[name]), [-1])
                    for name in self.sequential_names
                },
            )

        def _postprocess(contextual, sequential):
            sequential = {
                # Convert the sequential columns from dense to sparse
                name: self.schema[name].to_sparse(sequential[name])
                for name in self.sequential_names
            }
            return {**contextual, **sequential}

        def _shape():
            return (
                {name: tf.TensorShape([]) for name in self.contextual_names},
                {
                    name: tf.TensorShape([None])
                    for name in self.sequential_names
                },
            )

        # List all files matching a given pattern
        pattern = [self.path, name, 'examples', 'part-*']
        dataset = tf.data.Dataset.list_files(os.path.join(*pattern))
        # Shuffle the files if needed
        if 'shuffle_macro' in config:
            dataset = dataset.shuffle(**config['shuffle_macro'])
        # Convert the files into datasets of examples stored as TFRecords and
        # amalgamate these datasets into one dataset of examples
        dataset = dataset \
            .interleave(tf.data.TFRecordDataset, **config['interleave'])
        # Shuffle the examples if needed
        if 'shuffle_micro' in config:
            dataset = dataset.shuffle(**config['shuffle_micro'])
        # Preprocess the examples with respect to a given spec, pad the examples
        # and form batches of different sizes, and postprocess the batches
        if config.get('transformed', False):
            _preprocess = _preprocess_transformed
        else:
            _preprocess = _preprocess_untransformed
        dataset = dataset \
            .map(_preprocess, **config['map']) \
            .padded_batch(padded_shapes=_shape(), **config['batch']) \
            .map(_postprocess, **config['map'])
        # Prefetch the batches if needed
        if 'prefetch' in config:
            dataset = dataset.prefetch(**config['prefetch'])
        # Repeat the data once the source is exhausted if needed
        if 'repeat' in config:
            dataset = dataset.repeat(**config['repeat'])
        return dataset

    def create_feature_columns(self, scope):

        def _process(name):
            return self.schema[name].to_feature_column(self.transform)

        return list(map(_process, getattr(self, scope + '_feature_names')))



def timestamp(object, now):
    if isinstance(object, dict):
        return {name: timestamp(value, now) for name, value in object.items()}
    if isinstance(object, str):
        return now.strftime(object)
    return object


def list_files(path):
    from tensorflow.python.lib.io.file_io import get_matching_files
    return sorted(get_matching_files(path))



import apache_beam as beam
import os
import tensorflow_transform as tft

from functools import partial
from tensorflow_transform.beam import impl as tt_beam
from tensorflow_transform.beam.tft_beam_io import transform_fn_io
from tensorflow_transform.tf_metadata import dataset_metadata
from tensorflow_transform.tf_metadata import dataset_schema

from forecast.schema import Schema


class Pipeline(beam.Pipeline):

    def __init__(self, config):
        # Set options pertaining to the Dataflow runner
        options = dict(
            runner='DataflowRunner',
            staging_location=_locate(config, 'staging'),
            temp_location=_locate(config, 'temporary'),
            setup_file=os.path.join('.', 'setup.py'),
            save_main_session=True,
            flags=[],
        )
        options.update(config['pipeline'])
        # Create and populate a Beam pipeline
        super().__init__(options=beam.pipeline.PipelineOptions(**options))
        with tt_beam.Context(temp_dir=_locate(config, 'temporary')):
            _populate(self, config)


def _locate(config, *names):
    return os.path.join(config['output']['path'], *names)


def _populate(pipeline, config):
    # Create a schema according to the configuration file
    schema = Schema(config['data']['schema'])

    def _analyze(example):
        return {
            name: _analyze_column(schema[name], example[name])
            for name in example.keys()
        }

    def _analyze_column(column, value):
        if column.transform is None:
            return value
        if column.transform == 'z':
            return tft.scale_to_z_score(value)
        # Define other transforms, such as vocabulary look-up
        assert False

    def _filter(mode, example):
        return mode['name'] in example['mode']

    # Read the SQL code
    query = open(config['data']['path']).read()
    # Create a BigQuery source
    source = beam.io.BigQuerySource(query=query, use_standard_sql=True)
    # Create metadata needed later
    spec = schema.to_feature_spec()
    meta = dataset_metadata.DatasetMetadata(
        schema=dataset_schema.from_feature_spec(spec))
    # Read data from BigQuery
    data = pipeline \
        | 'read' >> beam.io.Read(source)
    # Loop over modes whose purpose is analysis
    transform_functions = {}
    for mode in config['modes']:
        if 'transform' in mode:
            continue
        name = mode['name']
        # Select examples that belong to the current mode
        data_ = data \
            | name + '-filter' >> beam.Filter(partial(_filter, mode))
        # Analyze the examples
        transform_functions[name] = (data_, meta) \
            | name + '-analyze' >> tt_beam.AnalyzeDataset(_analyze)
        path = _locate(config, name, 'transform')
        # Store the transform function
        transform_functions[name] \
            | name + '-write-transform' >> transform_fn_io.WriteTransformFn(path)
    # Loop over modes whose purpose is transformation
    for mode in config['modes']:
        if not 'transform' in mode:
            continue
        name = mode['name']
        # Select examples that belong to the current mode
        data_ = data \
            | name + '-filter' >> beam.Filter(partial(_filter, mode))
        # Shuffle examples if needed
        if mode.get('shuffle', False):
            data_ = data_ \
                | name + '-shuffle' >> beam.transforms.Reshuffle()
        # Transform the examples using an appropriate transform function
        if mode['transform'] == 'identity':
            coder = tft.coders.ExampleProtoCoder(meta.schema)
        else:
            data_, meta_ = ((data_, meta), transform_functions[mode['transform']]) \
                | name + '-transform' >> tt_beam.TransformDataset()
            coder = tft.coders.ExampleProtoCoder(meta_.schema)
        path = _locate(config, name, 'examples', 'part')
        # Store the transformed examples as TFRecords
        data_ \
            | name + '-encode' >> beam.Map(coder.encode) \
            | name + '-write-examples' >> beam.io.tfrecordio.WriteToTFRecord(path)



import tensorflow as tf

from collections import OrderedDict


class Field:

    def __init__(self, name, kind, transform=None):
        if isinstance(kind, list):
            self.scope = 'sequential'
            kind = getattr(tf, kind[0])
        else:
            self.scope = 'contextual'
            kind = getattr(tf, kind)
        self.name = name
        self.kind = kind
        self.transform = transform

    def to_dense(self, tensor):
        value = '-1' if self.kind == tf.string else -1
        return tf.sparse.to_dense(tensor, default_value=value)

    def to_sparse(self, tensor):
        value = '-1' if self.kind == tf.string else -1
        index = tf.where(tf.not_equal(tensor, value))
        return tf.SparseTensor(index, tf.gather_nd(tensor, index),
                               tf.shape(tensor, out_type=tf.int64))

    def to_feature_column(self, transform):
        if self.transform is None:
            if self.scope == 'sequential':
                function = tf.feature_column.sequence_numeric_column
            else:
                function = tf.feature_column.numeric_column
            return function(self.name)
        assert False


class Schema(OrderedDict):

    def __init__(self, fields):
        fields = list(map(lambda options: Field(**options), fields))
        names = map(lambda field: field.name, fields)
        super().__init__(zip(names, fields))

    def select(self, scope):
        return [name for name, field in self.items() if field.scope == scope]

    def to_feature_spec(self):

        def _process(name):
            if self[name].scope == 'sequential':
                return tf.io.VarLenFeature(self[name].kind)
            else:
                return tf.io.FixedLenFeature([], self[name].kind)

        return {name: _process(name) for name in self.keys()}



import argparse
import datetime
import json

from forecast.support import timestamp
from forecast.task import Task

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--action', type=str, required=True)
    parser.add_argument('--config', type=str, required=True)
    arguments = parser.parse_args()
    with open(arguments.config) as file:
        config = timestamp(json.load(file), datetime.datetime.now())
        Task().run(arguments.action, config)



#!/usr/bin/env python

from setuptools import find_packages
from setuptools import setup

setup(
    name='bigquery-pipeline',
    version='0.0.1',
    author='Antoine Delorme',
    author_email='antoine.delorme@gmail.com',
    packages=find_packages(),
    install_requires=[
        'google-cloud-core>=1.0.0',
        'tensorflow-metadata==0.14.0',
        'tensorflow-transform==0.14.0',
    ],
)
